@@ -12,7 +12,9 @@
 # implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
+from __future__ import unicode_literals
 import logging
+
 import os
 
 from concurrent.futures import as_completed, CancelledError, TimeoutError
@@ -23,13 +25,14 @@ from os import environ, makedirs, stat, utime
 from os.path import (
     basename, dirname, getmtime, getsize, isdir, join, sep as os_path_sep
 )
+from posixpath import join as urljoin
 from random import shuffle
 from time import time
 from threading import Thread
 from six import StringIO, text_type
 from six.moves.queue import Queue
 from six.moves.queue import Empty as QueueEmpty
-from six.moves.urllib.parse import quote, unquote
+from six.moves.urllib.parse import quote
 from six import Iterator, string_types
 
 import json
@@ -41,7 +44,7 @@ from swiftclient.command_helpers import (
 )
 from swiftclient.utils import (
     config_true_value, ReadableToIterable, LengthWrapper, EMPTY_ETAG,
-    parse_api_response, report_traceback
+    parse_api_response, report_traceback, n_groups
 )
 from swiftclient.exceptions import ClientException
 from swiftclient.multithreading import MultiThreadingManager
@@ -92,7 +95,7 @@ def process_options(options):
     # Use new-style args if old ones not present
     if not options['auth'] and options['os_auth_url']:
         options['auth'] = options['os_auth_url']
-    if not options['user']and options['os_username']:
+    if not options['user'] and options['os_username']:
         options['user'] = options['os_username']
     if not options['key'] and options['os_password']:
         options['key'] = options['os_password']
@@ -286,6 +289,7 @@ class SwiftUploadObject(object):
         if not self.object_name:
             raise SwiftError('Object names must not be empty strings')
 
+        self.object_name = self.object_name.lstrip('/')
         self.options = options
         self.source = source
 
@@ -380,6 +384,7 @@ class SwiftService(object):
             object_uu_threads=self._options['object_uu_threads'],
             container_threads=self._options['container_threads']
         )
+        self.capabilities_cache = {}  # Each instance should have its own cache
 
     def __enter__(self):
         self.thread_manager.__enter__()
@@ -999,7 +1004,7 @@ class SwiftService(object):
                         raise
                     raise SwiftError('Account not found', exc=err)
 
-        elif not objects:
+        elif objects is None:
             if '/' in container:
                 raise SwiftError('\'/\' in container name',
                                  container=container)
@@ -1159,11 +1164,15 @@ class SwiftService(object):
                 if fp is not None:
                     fp.close()
                     if 'x-object-meta-mtime' in headers and not no_file:
-                        mtime = float(headers['x-object-meta-mtime'])
-                        if options['out_file']:
-                            utime(options['out_file'], (mtime, mtime))
+                        try:
+                            mtime = float(headers['x-object-meta-mtime'])
+                        except ValueError:
+                            pass  # no real harm; couldn't trust it anyway
                         else:
-                            utime(path, (mtime, mtime))
+                            if options['out_file']:
+                                utime(options['out_file'], (mtime, mtime))
+                            else:
+                                utime(path, (mtime, mtime))
 
             res = {
                 'action': 'download_object',
@@ -1281,7 +1290,8 @@ class SwiftService(object):
         """
         Upload a list of objects to a given container.
 
-        :param container: The container to put the uploads into.
+        :param container: The container (or pseudo-folder path) to put the
+                          uploads into.
         :param objects: A list of file/directory names (strings) or
                         SwiftUploadObject instances containing a source for the
                         created object, an object name, and an options dict
@@ -1298,10 +1308,10 @@ class SwiftService(object):
 
                         The SwiftUploadObject source may be one of:
 
-                            file - A file like object (with a read method)
-                            path - A string containing the path to a local file
-                                   or directory
-                            None - Indicates that we want an empty object
+                            * A file-like object (with a read method)
+                            * A string containing the path to a local
+                              file or directory
+                            * None, to indicate that we want an empty object
 
         :param options: A dictionary containing options to override the global
                         options specified during the service object creation.
@@ -1338,6 +1348,11 @@ class SwiftService(object):
         except ValueError:
             raise SwiftError('Segment size should be an integer value')
 
+        # Incase we have a psudeo-folder path for <container> arg, derive
+        # the container name from the top path and prepend the rest to
+        # the object name. (same as passing --object-name).
+        container, _sep, pseudo_folder = container.partition('/')
+
         # Try to create the container, just in case it doesn't exist. If this
         # fails, it might just be because the user doesn't have container PUT
         # permissions, so we'll ignore any error. If there's really a problem,
@@ -1349,8 +1364,7 @@ class SwiftService(object):
                 _header[POLICY]
         create_containers = [
             self.thread_manager.container_pool.submit(
-                self._create_container_job, container, headers=policy_header
-            )
+                self._create_container_job, container, headers=policy_header)
         ]
 
         # wait for first container job to complete before possibly attempting
@@ -1394,7 +1408,7 @@ class SwiftService(object):
         rq = Queue()
         file_jobs = {}
 
-        upload_objects = self._make_upload_objects(objects)
+        upload_objects = self._make_upload_objects(objects, pseudo_folder)
         for upload_object in upload_objects:
             s = upload_object.source
             o = upload_object.object_name
@@ -1485,14 +1499,16 @@ class SwiftService(object):
             res = get_from_queue(rq)
 
     @staticmethod
-    def _make_upload_objects(objects):
+    def _make_upload_objects(objects, pseudo_folder=''):
         upload_objects = []
 
         for o in objects:
             if isinstance(o, string_types):
-                obj = SwiftUploadObject(o)
+                obj = SwiftUploadObject(o, urljoin(pseudo_folder,
+                                                   o.lstrip('/')))
                 upload_objects.append(obj)
             elif isinstance(o, SwiftUploadObject):
+                o.object_name = urljoin(pseudo_folder, o.object_name)
                 upload_objects.append(o)
             else:
                 raise SwiftError(
@@ -1628,6 +1644,7 @@ class SwiftService(object):
 
         res = {
             'action': 'upload_segment',
+            'for_container': container,
             'for_object': obj_name,
             'segment_index': segment_index,
             'segment_size': segment_size,
@@ -1908,7 +1925,8 @@ class SwiftService(object):
                     res['manifest_response_dict'] = mr
                 else:
                     new_object_manifest = '%s/%s/%s/%s/%s/' % (
-                        quote(seg_container), quote(obj),
+                        quote(seg_container.encode('utf8')),
+                        quote(obj.encode('utf8')),
                         put_headers['x-object-meta-mtime'], full_size,
                         options['segment_size'])
                     if old_manifest and old_manifest.rstrip('/') == \
@@ -1954,8 +1972,7 @@ class SwiftService(object):
                 delobjsmap = {}
                 if old_manifest:
                     scontainer, sprefix = old_manifest.split('/', 1)
-                    scontainer = unquote(scontainer)
-                    sprefix = unquote(sprefix).rstrip('/') + '/'
+                    sprefix = sprefix.rstrip('/') + '/'
                     delobjsmap[scontainer] = []
                     for part in self.list(scontainer, {'prefix': sprefix}):
                         if not part["success"]:
@@ -2032,13 +2049,14 @@ class SwiftService(object):
                             {
                                 'yes_all': False,
                                 'leave_segments': False,
+                                'prefix': None,
                             }
 
         :returns: A generator for returning the results of the delete
                   operations. Each result yielded from the generator is either
-                  a 'delete_container', 'delete_object' or 'delete_segment'
-                  dictionary containing the results of an individual delete
-                  operation.
+                  a 'delete_container', 'delete_object', 'delete_segment', or
+                  'bulk_delete' dictionary containing the results of an
+                  individual delete operation.
 
         :raises: ClientException
         :raises: SwiftError
@@ -2048,19 +2066,24 @@ class SwiftService(object):
         else:
             options = self._options
 
-        rq = Queue()
         if container is not None:
             if objects is not None:
+                if options['prefix']:
+                    objects = [obj for obj in objects
+                               if obj.startswith(options['prefix'])]
+                rq = Queue()
                 obj_dels = {}
-                for obj in objects:
-                    obj_del = self.thread_manager.object_dd_pool.submit(
-                        self._delete_object, container, obj, options,
-                        results_queue=rq
-                    )
-                    obj_details = {'container': container, 'object': obj}
-                    obj_dels[obj_del] = obj_details
 
-                # Start a thread to watch for upload results
+                if self._should_bulk_delete(objects):
+                    for obj_slice in n_groups(
+                            objects, self._options['object_dd_threads']):
+                        self._bulk_delete(container, obj_slice, options,
+                                          obj_dels)
+                else:
+                    self._per_item_delete(container, objects, options,
+                                          obj_dels, rq)
+
+                # Start a thread to watch for delete results
                 Thread(
                     target=self._watch_futures, args=(obj_dels, rq)
                 ).start()
@@ -2083,6 +2106,8 @@ class SwiftService(object):
         else:
             if objects:
                 raise SwiftError('Objects specified without container')
+            if options['prefix']:
+                raise SwiftError('Prefix specified without container')
             if options['yes_all']:
                 cancelled = False
                 containers = []
@@ -2106,6 +2131,33 @@ class SwiftService(object):
                                     and not res['success']):
                                 cancelled = True
 
+    def _should_bulk_delete(self, objects):
+        if len(objects) < 2 * self._options['object_dd_threads']:
+            # Not many objects; may as well delete one-by-one
+            return False
+
+        try:
+            cap_result = self.capabilities()
+            if not cap_result['success']:
+                # This shouldn't actually happen, but just in case we start
+                # being more nuanced about our capabilities result...
+                return False
+        except ClientException:
+            # Old swift, presumably; assume no bulk middleware
+            return False
+
+        swift_info = cap_result['capabilities']
+        return 'bulk_delete' in swift_info
+
+    def _per_item_delete(self, container, objects, options, rdict, rq):
+        for obj in objects:
+            obj_del = self.thread_manager.object_dd_pool.submit(
+                self._delete_object, container, obj, options,
+                results_queue=rq
+            )
+            obj_details = {'container': container, 'object': obj}
+            rdict[obj_del] = obj_details
+
     @staticmethod
     def _delete_segment(conn, container, obj, results_queue=None):
         results_dict = {}
@@ -2164,8 +2216,7 @@ class SwiftService(object):
                 dlo_segments_deleted = True
                 segment_pool = self.thread_manager.segment_pool
                 s_container, s_prefix = old_manifest.split('/', 1)
-                s_container = unquote(s_container)
-                s_prefix = unquote(s_prefix).rstrip('/') + '/'
+                s_prefix = s_prefix.rstrip('/') + '/'
 
                 del_segs = []
                 for part in self.list(
@@ -2235,18 +2286,20 @@ class SwiftService(object):
 
     def _delete_container(self, container, options):
         try:
-            for part in self.list(container=container):
-                if part["success"]:
-                    objs = [o['name'] for o in part['listing']]
+            for part in self.list(container=container, options=options):
+                if not part["success"]:
 
-                    o_dels = self.delete(
-                        container=container, objects=objs, options=options
-                    )
-                    for res in o_dels:
-                        yield res
-                else:
                     raise part["error"]
 
+                for res in self.delete(
+                        container=container,
+                        objects=[o['name'] for o in part['listing']],
+                        options=options):
+                    yield res
+            if options['prefix']:
+                # We're only deleting a subset of objects within the container
+                return
+
             con_del = self.thread_manager.container_pool.submit(
                 self._delete_empty_container, container
             )
@@ -2267,9 +2320,55 @@ class SwiftService(object):
 
         yield con_del_res
 
+    # Bulk methods
+    #
+    def _bulk_delete(self, container, objects, options, rdict):
+        if objects:
+            bulk_del = self.thread_manager.object_dd_pool.submit(
+                self._bulkdelete, container, objects, options
+            )
+            bulk_details = {'container': container, 'objects': objects}
+            rdict[bulk_del] = bulk_details
+
+    @staticmethod
+    def _bulkdelete(conn, container, objects, options):
+        results_dict = {}
+        try:
+            headers = {
+                'Accept': 'application/json',
+                'Content-Type': 'text/plain',
+            }
+            res = {'container': container, 'objects': objects}
+            objects = [quote(('/%s/%s' % (container, obj)).encode('utf-8'))
+                       for obj in objects]
+            headers, body = conn.post_account(
+                headers=headers,
+                query_string='bulk-delete',
+                data=b''.join(obj.encode('utf-8') + b'\n' for obj in objects),
+                response_dict=results_dict)
+            if body:
+                res.update({'success': True,
+                            'result': parse_api_response(headers, body)})
+            else:
+                res.update({
+                    'success': False,
+                    'error': SwiftError(
+                        'No content received on account POST. '
+                        'Is the bulk operations middleware enabled?')})
+        except Exception as e:
+            res.update({'success': False, 'error': e})
+
+        res.update({
+            'action': 'bulk_delete',
+            'attempts': conn.attempts,
+            'response_dict': results_dict
+        })
+
+        return res
+
     # Capabilities related methods
     #
-    def capabilities(self, url=None):
+    def capabilities(self, url=None, refresh_cache=False):
         """
         List the cluster capabilities.
 
@@ -2278,30 +2377,29 @@ class SwiftService(object):
         :returns: A dictionary containing the capabilities of the cluster.
 
         :raises: ClientException
-        :raises: SwiftError
         """
+        if not refresh_cache and url in self.capabilities_cache:
+            return self.capabilities_cache[url]
+
         res = {
-            'action': 'capabilities'
+            'action': 'capabilities',
+            'timestamp': time(),
         }
 
-        try:
-            cap = self.thread_manager.container_pool.submit(
-                self._get_capabilities, url
-            )
-            capabilities = get_future_result(cap)
+        cap = self.thread_manager.container_pool.submit(
+            self._get_capabilities, url
+        )
+        capabilities = get_future_result(cap)
+        res.update({
+            'success': True,
+            'capabilities': capabilities
+        })
+        if url is not None:
             res.update({
-                'success': True,
-                'capabilities': capabilities
+                'url': url
             })
-            if url is not None:
-                res.update({
-                    'url': url
-                })
-        except ClientException as err:
-            if err.http_status != 404:
-                raise err
-            raise SwiftError('Account not found', exc=err)
 
+        self.capabilities_cache[url] = res
         return res
 
     @staticmethod
