@@ -23,7 +23,8 @@ import socket
 from optparse import OptionParser, OptionGroup, SUPPRESS_HELP
 from os import environ, walk, _exit as os_exit
 from os.path import isfile, isdir, join
-from six import text_type
+from six import text_type, PY2
+from six.moves.urllib.parse import unquote
 from sys import argv as sys_argv, exit, stderr
 from time import gmtime, strftime
 
@@ -32,6 +33,7 @@ from swiftclient.utils import config_true_value, generate_temp_url, prt_bytes
 from swiftclient.multithreading import OutputManager
 from swiftclient.exceptions import ClientException
 from swiftclient import __version__ as client_version
+from swiftclient.client import logger_settings as client_logger_settings
 from swiftclient.service import SwiftService, SwiftError, \
     SwiftUploadObject, get_conn
 from swiftclient.command_helpers import print_account_stats, \
@@ -51,18 +53,18 @@ def immediate_exit(signum, frame):
     stderr.write(" Aborted\n")
     os_exit(2)
 
-st_delete_options = '''[-all] [--leave-segments]
+st_delete_options = '''[--all] [--leave-segments]
                     [--object-threads <threads>]
                     [--container-threads <threads>]
-                    <container> [object]
+                    [<container>] [<object>] [...]
 '''
 
 st_delete_help = '''
 Delete a container or objects within a container.
 
 Positional arguments:
-  <container>           Name of container to delete from.
-  [object]              Name of object to delete. Specify multiple times
+  [<container>]         Name of container to delete from.
+  [<object>]            Name of object to delete. Specify multiple times
                         for multiple objects.
 
 Optional arguments:
@@ -81,6 +83,9 @@ def st_delete(parser, args, output_manager):
     parser.add_option(
         '-a', '--all', action='store_true', dest='yes_all',
         default=False, help='Delete all containers and objects.')
+    parser.add_option(
+        '-p', '--prefix', dest='prefix',
+        help='Only delete items beginning with the <prefix>.')
     parser.add_option(
         '', '--leave-segments', action='store_true',
         dest='leave_segments', default=False,
@@ -128,25 +133,55 @@ def st_delete(parser, args, output_manager):
                 o = r.get('object', '')
                 a = r.get('attempts')
 
-                if r['success']:
-                    if options.verbose:
-                        a = ' [after {0} attempts]'.format(a) if a > 1 else ''
+                if r['action'] == 'bulk_delete':
+                    if r['success']:
+                        objs = r.get('objects', [])
+                        for o, err in r.get('result', {}).get('Errors', []):
+                            # o will be of the form quote("/<cont>/<obj>")
+                            o = unquote(o)
+                            if PY2:
+                                # In PY3, unquote(unicode) uses utf-8 like we
+                                # want, but PY2 uses latin-1
+                                o = o.encode('latin-1').decode('utf-8')
+                            output_manager.error('Error Deleting: {0}: {1}'
+                                                 .format(o[1:], err))
+                            try:
+                                objs.remove(o[len(c) + 2:])
+                            except ValueError:
+                                # shouldn't happen, but ignoring it won't hurt
+                                pass
 
-                        if r['action'] == 'delete_object':
+                        for o in objs:
                             if options.yes_all:
                                 p = '{0}/{1}'.format(c, o)
                             else:
                                 p = o
-                        elif r['action'] == 'delete_segment':
-                            p = '{0}/{1}'.format(c, o)
-                        elif r['action'] == 'delete_container':
-                            p = c
-
-                        output_manager.print_msg('{0}{1}'.format(p, a))
+                            output_manager.print_msg('{0}{1}'.format(p, a))
+                    else:
+                        for o in r.get('objects', []):
+                            output_manager.error('Error Deleting: {0}/{1}: {2}'
+                                                 .format(c, o, r['error']))
                 else:
-                    p = '{0}/{1}'.format(c, o) if o else c
-                    output_manager.error('Error Deleting: {0}: {1}'
-                                         .format(p, r['error']))
+                    if r['success']:
+                        if options.verbose:
+                            a = (' [after {0} attempts]'.format(a)
+                                 if a > 1 else '')
+
+                            if r['action'] == 'delete_object':
+                                if options.yes_all:
+                                    p = '{0}/{1}'.format(c, o)
+                                else:
+                                    p = o
+                            elif r['action'] == 'delete_segment':
+                                p = '{0}/{1}'.format(c, o)
+                            elif r['action'] == 'delete_container':
+                                p = c
+
+                            output_manager.print_msg('{0}{1}'.format(p, a))
+                    else:
+                        p = '{0}/{1}'.format(c, o) if o else c
+                        output_manager.error('Error Deleting: {0}: {1}'
+                                             .format(p, r['error']))
         except SwiftError as err:
             output_manager.error(err.value)
 
@@ -156,7 +191,7 @@ st_download_options = '''[--all] [--marker] [--prefix <prefix>]
                       [--object-threads <threads>]
                       [--container-threads <threads>] [--no-download]
                       [--skip-identical] [--remove-prefix]
-                      [--header <header:value>]
+                      [--header <header:value>] [--no-shuffle]
                       <container> <object>
 '''
 
@@ -376,7 +411,7 @@ def st_download(parser, args, output_manager):
 
 
 st_list_options = '''[--long] [--lh] [--totals] [--prefix <prefix>]
-                  [--delimiter <delimiter>]
+                  [--delimiter <delimiter>] [container]
 '''
 
 st_list_help = '''
@@ -390,8 +425,10 @@ Optional arguments:
   --lh                  Report sizes in human readable format similar to
                         ls -lh.
   -t, --totals          Used with -l or --lh, only report totals.
-  -p, --prefix          Only list items beginning with the prefix.
-  -d, --delimiter       Roll up items with the given delimiter. For containers
+  -p <prefix>, --prefix <prefix>
+                        Only list items beginning with the prefix.
+  -d <delim>, --delimiter <delim>
+                        Roll up items with the given delimiter. For containers
                         only. See OpenStack Swift API documentation for what
                         this means.
 '''.strip('\n')
@@ -424,6 +461,7 @@ def st_list(parser, args, output_manager):
                             datestamp, item_name)
                 else:    # list container contents
                     subdir = item.get('subdir')
+                    content_type = item.get('content_type')
                     if subdir is None:
                         item_bytes = item.get('bytes')
                         byte_str = prt_bytes(item_bytes, options.human)
@@ -436,7 +474,8 @@ def st_list(parser, args, output_manager):
                         item_name = subdir
                     if not options.totals:
                         output_manager.print_msg(
-                            "%s %10s %8s %s", byte_str, date, xtime, item_name)
+                            "%s %10s %8s %24s %s",
+                            byte_str, date, xtime, content_type, item_name)
                 total_bytes += item_bytes
 
         # report totals
@@ -688,7 +727,7 @@ st_upload_options = '''[--changed] [--skip-identical] [--segment-size <size>]
                     [--object-threads <thread>] [--segment-threads <threads>]
                     [--header <header>] [--use-slo] [--ignore-checksum]
                     [--object-name <object-name>]
-                    <container> <file_or_directory>
+                    <container> <file_or_directory> [<file_or_directory>] [...]
 '''
 
 st_upload_help = ''' Uploads specified files and directories to the given container.
@@ -893,13 +932,14 @@ def st_upload(parser, args, output_manager):
                                 if error.http_response_content:
                                     if msg:
                                         msg += ': '
-                                    msg += error.http_response_content[:60]
+                                    msg += (error.http_response_content
+                                            .decode('utf8')[:60])
                                 msg = ': %s' % msg
                         else:
                             msg = ': %s' % error
                         output_manager.warning(
                             'Warning: failed to create container '
-                            "'%s'%s", container, msg
+                            "'%s'%s", r['container'], msg
                         )
                     else:
                         output_manager.error("%s" % error)
@@ -1004,7 +1044,8 @@ def st_auth(parser, args, thread_manager):
         print('export OS_AUTH_TOKEN=%s' % sh_quote(token))
 
 
-st_tempurl_options = '<method> <seconds> <path> <key>'
+st_tempurl_options = '''[--absolute]
+                     <method> <seconds> <path> <key>'''
 
 
 st_tempurl_help = '''
@@ -1023,7 +1064,7 @@ Positional arguments:
                         "Temp-URL-Key:b3968d0207b54ece87cccc06515a89d4"\'
 
 Optional arguments:
-  --absolute            Interpet the <seconds> positional argument as a Unix
+  --absolute            Interpret the <seconds> positional argument as a Unix
                         timestamp rather than a number of seconds in the
                         future.
 '''.strip('\n')
@@ -1062,6 +1103,14 @@ def parse_args(parser, args, enforce_requires=True):
     if not args:
         args = ['-h']
     (options, args) = parser.parse_args(args)
+    if enforce_requires and (options.debug or options.info):
+        logging.getLogger("swiftclient")
+        if options.debug:
+            logging.basicConfig(level=logging.DEBUG)
+            logging.getLogger('iso8601').setLevel(logging.WARNING)
+            client_logger_settings['redact_sensitive_headers'] = False
+        elif options.info:
+            logging.basicConfig(level=logging.INFO)
 
     if len(args) > 1 and args[1] == '--help':
         _help = globals().get('st_%s_help' % args[0],
@@ -1151,9 +1200,11 @@ def main(arguments=None):
     version = client_version
     parser = OptionParser(version='python-swiftclient %s' % version,
                           usage='''
-usage: %%prog [--version] [--help] [--os-help] [--snet] [--verbose]
+usage: %prog [--version] [--help] [--os-help] [--snet] [--verbose]
              [--debug] [--info] [--quiet] [--auth <auth_url>]
-             [--auth-version <auth_version>] [--user <username>]
+             [--auth-version <auth_version> |
+                 --os-identity-api-version <auth_version> ]
+             [--user <username>]
              [--key <api_key>] [--retries <num_retries>]
              [--os-username <auth-user-name>] [--os-password <auth-password>]
              [--os-user-id <auth-user-id>]
@@ -1191,29 +1242,38 @@ Positional arguments:
     auth                 Display auth related environment variables.
 
 Examples:
-  %%prog download --help
+  %prog download --help
 
-  %%prog -A https://auth.api.rackspacecloud.com/v1.0 -U user -K api_key stat -v
+  %prog -A https://auth.api.rackspacecloud.com/v1.0 -U user -K api_key stat -v
 
-  %%prog --os-auth-url https://api.example.com/v2.0 --os-tenant-name tenant \\
+  %prog --os-auth-url https://api.example.com/v2.0 --os-tenant-name tenant \\
       --os-username user --os-password password list
 
-  %%prog --os-auth-url https://api.example.com/v3 --auth-version 3\\
+  %prog --os-auth-url https://api.example.com/v3 --auth-version 3\\
       --os-project-name project1 --os-project-domain-name domain1 \\
       --os-username user --os-user-domain-name domain1 \\
       --os-password password list
 
-  %%prog --os-auth-url https://api.example.com/v3 --auth-version 3\\
+  %prog --os-auth-url https://api.example.com/v3 --auth-version 3\\
       --os-project-id 0123456789abcdef0123456789abcdef \\
       --os-user-id abcdef0123456789abcdef0123456789 \\
       --os-password password list
 
-  %%prog --os-auth-token 6ee5eb33efad4e45ab46806eac010566 \\
+  %prog --os-auth-token 6ee5eb33efad4e45ab46806eac010566 \\
       --os-storage-url https://10.1.5.2:8080/v1/AUTH_ced809b6a4baea7aeab61a \\
       list
 
-  %%prog list --lh
-'''.strip('\n') % globals())
+  %prog list --lh
+'''.strip('\n'))
+
+    default_auth_version = '1.0'
+    for k in ('ST_AUTH_VERSION', 'OS_AUTH_VERSION', 'OS_IDENTITY_API_VERSION'):
+        try:
+            default_auth_version = environ[k]
+            break
+        except KeyError:
+            pass
+
     parser.add_option('--os-help', action='store_true', dest='os_help',
                       help='Show OpenStack authentication options.')
     parser.add_option('--os_help', action='store_true', help=SUPPRESS_HELP)
@@ -1232,14 +1292,14 @@ Examples:
     parser.add_option('-A', '--auth', dest='auth',
                       default=environ.get('ST_AUTH'),
                       help='URL for obtaining an auth token.')
-    parser.add_option('-V', '--auth-version',
+    parser.add_option('-V', '--auth-version', '--os-identity-api-version',
                       dest='auth_version',
-                      default=environ.get('ST_AUTH_VERSION',
-                                          (environ.get('OS_AUTH_VERSION',
-                                                       '1.0'))),
+                      default=default_auth_version,
                       type=str,
                       help='Specify a version for authentication. '
-                           'Defaults to 1.0.')
+                           'Defaults to env[ST_AUTH_VERSION], '
+                           'env[OS_AUTH_VERSION], env[OS_IDENTITY_API_VERSION]'
+                           ' or 1.0.')
     parser.add_option('-U', '--user', dest='user',
                       default=environ.get('ST_USER'),
                       help='User name for obtaining an auth token.')
@@ -1410,13 +1470,6 @@ Examples:
 
     signal.signal(signal.SIGINT, immediate_exit)
 
-    if options.debug or options.info:
-        logging.getLogger("swiftclient")
-        if options.debug:
-            logging.basicConfig(level=logging.DEBUG)
-        elif options.info:
-            logging.basicConfig(level=logging.INFO)
-
     with OutputManager() as output:
 
         parser.usage = globals()['st_%s_help' % args[0]]
