@@ -18,7 +18,6 @@ from requests import RequestException
 from requests.structures import CaseInsensitiveDict
 from time import sleep
 import unittest
-import testtools
 import mock
 import six
 from six.moves import reload_module
@@ -100,31 +99,15 @@ def fake_http_connect(*code_iter, **kwargs):
             self._is_closed = True
             self.headers = headers or {}
 
-        def connect(self):
-            self._is_closed = False
-
-        def close(self):
-            self._is_closed = True
-
-        def isclosed(self):
-            return self._is_closed
-
         def getresponse(self):
             if kwargs.get('raise_exc'):
                 raise Exception('test')
             return self
 
-        def getexpect(self):
-            if self.status == -2:
-                raise RequestException()
-            if self.status == -3:
-                return FakeConn(507)
-            return FakeConn(100)
-
         def getheaders(self):
             if self.headers:
                 return self.headers.items()
-            headers = {'content-length': len(self.body),
+            headers = {'content-length': str(len(self.body)),
                        'content-type': 'x-application/test',
                        'x-timestamp': self.timestamp,
                        'last-modified': self.timestamp,
@@ -132,7 +115,7 @@ def fake_http_connect(*code_iter, **kwargs):
                        'etag':
                        self.etag or '"%s"' % EMPTY_ETAG,
                        'x-works': 'yes',
-                       'x-account-container-count': 12345}
+                       'x-account-container-count': '12345'}
             if not self.timestamp:
                 del headers['x-timestamp']
             try:
@@ -199,21 +182,20 @@ def fake_http_connect(*code_iter, **kwargs):
                                  timestamp=timestamp)
         if fake_conn.status <= 0:
             raise RequestException()
-        fake_conn.connect()
         return fake_conn
 
     connect.code_iter = code_iter
     return connect
 
 
-class MockHttpTest(testtools.TestCase):
+class MockHttpTest(unittest.TestCase):
 
     def setUp(self):
         super(MockHttpTest, self).setUp()
         self.fake_connect = None
         self.request_log = []
 
-        # Capture output, since the test-runner stdout/stderr moneky-patching
+        # Capture output, since the test-runner stdout/stderr monkey-patching
         # won't cover the references to sys.stdout/sys.stderr in
         # swiftclient.multithreading
         self.capture_output = CaptureOutput()
@@ -250,7 +232,6 @@ class MockHttpTest(testtools.TestCase):
                     self.request_log.append((parsed, method, url, args,
                                              kwargs, conn.resp))
                     conn.host = conn.resp.host
-                    conn.isclosed = conn.resp.isclosed
                     conn.resp.has_been_read = False
                     _orig_read = conn.resp.read
 
@@ -263,9 +244,8 @@ class MockHttpTest(testtools.TestCase):
                         conn.resp.status = status
                     if auth_token:
                         headers = args[1]
-                        self.assertTrue('X-Auth-Token' in headers)
-                        actual_token = headers.get('X-Auth-Token')
-                        self.assertEqual(auth_token, actual_token)
+                        self.assertEqual(auth_token,
+                                         headers.get('X-Auth-Token'))
                     if query_string:
                         self.assertTrue(url.endswith('?' + query_string))
                     if url.endswith('invalid_cert') and not insecure:
@@ -325,7 +305,7 @@ class MockHttpTest(testtools.TestCase):
             self.orig_assertEqual(body, real_request['body'], err_msg)
 
         if len(expected) > 3:
-            headers = expected[3]
+            headers = CaseInsensitiveDict(expected[3])
             for key, value in headers.items():
                 real_request['key'] = key
                 real_request['expected_value'] = value
@@ -336,16 +316,30 @@ class MockHttpTest(testtools.TestCase):
                     'for %(method)s %(path)s %(headers)r' % real_request)
                 self.orig_assertEqual(value, real_request['value'],
                                       err_msg)
+            real_request['extra_headers'] = dict(
+                (key, value) for key, value in real_request['headers'].items()
+                if key not in headers)
+            if real_request['extra_headers']:
+                self.fail('Received unexpected headers for %(method)s '
+                          '%(path)s, got %(extra_headers)r' % real_request)
 
     def assertRequests(self, expected_requests):
         """
         Make sure some requests were made like you expected, provide a list of
         expected requests, typically in the form of [(method, path), ...]
+        or [(method, path, body, headers), ...]
         """
         real_requests = self.iter_request_log()
         for expected in expected_requests:
             real_request = next(real_requests)
             self.assert_request_equal(expected, real_request)
+        try:
+            real_request = next(real_requests)
+        except StopIteration:
+            pass
+        else:
+            self.fail('At least one extra request received: %r' %
+                      real_request)
 
     def assert_request(self, expected_request):
         """
@@ -504,8 +498,8 @@ class FakeKeystone(object):
         self.token = token
 
     class _Client(object):
-        def __init__(self, endpoint, token, **kwargs):
-            self.auth_token = token
+        def __init__(self, endpoint, auth_token, **kwargs):
+            self.auth_token = auth_token
             self.endpoint = endpoint
             self.service_catalog = self.ServiceCatalog(endpoint)
 
@@ -520,8 +514,8 @@ class FakeKeystone(object):
 
     def Client(self, **kwargs):
         self.calls.append(kwargs)
-        self.client = self._Client(endpoint=self.endpoint, token=self.token,
-                                   **kwargs)
+        self.client = self._Client(
+            endpoint=self.endpoint, auth_token=self.token, **kwargs)
         return self.client
 
     class Unauthorized(Exception):
